@@ -1,7 +1,11 @@
-import sys
+import os
 from threading import Lock
 import time
 import types
+from typing import (
+    Any, Callable, Dict, Iterable, List, Optional, Sequence, Type, TypeVar,
+    Union,
+)
 
 from . import values  # retain this import style for testability
 from .context_managers import ExceptionCounter, InprogressTracker, Timer
@@ -9,15 +13,12 @@ from .metrics_core import (
     Metric, METRIC_LABEL_NAME_RE, METRIC_NAME_RE,
     RESERVED_METRIC_LABEL_NAME_RE,
 )
-from .registry import REGISTRY
+from .registry import Collector, CollectorRegistry, REGISTRY
+from .samples import Exemplar, Sample
 from .utils import floatToGoString, INF
 
-if sys.version_info > (3,):
-    unicode = str
-    create_bound_method = types.MethodType
-else:
-    def create_bound_method(func, obj):
-        return types.MethodType(func, obj, obj.__class__)
+T = TypeVar('T', bound='MetricWrapperBase')
+F = TypeVar("F", bound=Callable[..., Any])
 
 
 def _build_full_name(metric_type, name, namespace, subsystem, unit):
@@ -36,21 +37,42 @@ def _build_full_name(metric_type, name, namespace, subsystem, unit):
     return full_name
 
 
+def _validate_labelname(l):
+    if not METRIC_LABEL_NAME_RE.match(l):
+        raise ValueError('Invalid label metric name: ' + l)
+    if RESERVED_METRIC_LABEL_NAME_RE.match(l):
+        raise ValueError('Reserved label metric name: ' + l)
+
+
 def _validate_labelnames(cls, labelnames):
     labelnames = tuple(labelnames)
     for l in labelnames:
-        if not METRIC_LABEL_NAME_RE.match(l):
-            raise ValueError('Invalid label metric name: ' + l)
-        if RESERVED_METRIC_LABEL_NAME_RE.match(l):
-            raise ValueError('Reserved label metric name: ' + l)
+        _validate_labelname(l)
         if l in cls._reserved_labelnames:
             raise ValueError('Reserved label metric name: ' + l)
     return labelnames
 
 
-class MetricWrapperBase(object):
-    _type = None
-    _reserved_labelnames = ()
+def _validate_exemplar(exemplar):
+    runes = 0
+    for k, v in exemplar.items():
+        _validate_labelname(k)
+        runes += len(k)
+        runes += len(v)
+    if runes > 128:
+        raise ValueError('Exemplar labels have %d UTF-8 characters, exceeding the limit of 128')
+
+
+def _get_use_created() -> bool:
+    return os.environ.get("PROMETHEUS_DISABLE_CREATED_SERIES", 'False').lower() not in ('true', '1', 't')
+
+
+_use_created = _get_use_created()
+
+
+class MetricWrapperBase(Collector):
+    _type: Optional[str] = None
+    _reserved_labelnames: Sequence[str] = ()
 
     def _is_observable(self):
         # Whether this metric is observable, i.e.
@@ -71,36 +93,36 @@ class MetricWrapperBase(object):
     def _get_metric(self):
         return Metric(self._name, self._documentation, self._type, self._unit)
 
-    def describe(self):
+    def describe(self) -> Iterable[Metric]:
         return [self._get_metric()]
 
-    def collect(self):
+    def collect(self) -> Iterable[Metric]:
         metric = self._get_metric()
-        for suffix, labels, value in self._samples():
-            metric.add_sample(self._name + suffix, labels, value)
+        for suffix, labels, value, timestamp, exemplar in self._samples():
+            metric.add_sample(self._name + suffix, labels, value, timestamp, exemplar)
         return [metric]
 
-    def __str__(self):
-        return "{0}:{1}".format(self._type, self._name)
+    def __str__(self) -> str:
+        return f"{self._type}:{self._name}"
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         metric_type = type(self)
-        return "{0}.{1}({2})".format(metric_type.__module__, metric_type.__name__, self._name)
+        return f"{metric_type.__module__}.{metric_type.__name__}({self._name})"
 
-    def __init__(self,
-                 name,
-                 documentation,
-                 labelnames=(),
-                 namespace='',
-                 subsystem='',
-                 unit='',
-                 registry=REGISTRY,
-                 _labelvalues=None,
-                 ):
+    def __init__(self: T,
+                 name: str,
+                 documentation: str,
+                 labelnames: Iterable[str] = (),
+                 namespace: str = '',
+                 subsystem: str = '',
+                 unit: str = '',
+                 registry: Optional[CollectorRegistry] = REGISTRY,
+                 _labelvalues: Optional[Sequence[str]] = None,
+                 ) -> None:
         self._name = _build_full_name(self._type, name, namespace, subsystem, unit)
         self._labelnames = _validate_labelnames(self, labelnames)
         self._labelvalues = tuple(_labelvalues or ())
-        self._kwargs = {}
+        self._kwargs: Dict[str, Any] = {}
         self._documentation = documentation
         self._unit = unit
 
@@ -110,7 +132,7 @@ class MetricWrapperBase(object):
         if self._is_parent():
             # Prepare the fields needed for child metrics.
             self._lock = Lock()
-            self._metrics = {}
+            self._metrics: Dict[Sequence[str], T] = {}
 
         if self._is_observable():
             self._metric_init()
@@ -120,7 +142,7 @@ class MetricWrapperBase(object):
             if registry:
                 registry.register(self)
 
-    def labels(self, *labelvalues, **labelkwargs):
+    def labels(self: T, *labelvalues: Any, **labelkwargs: Any) -> T:
         """Return the child for the given labelset.
 
         All metrics can have labels, allowing grouping of related time series.
@@ -147,7 +169,7 @@ class MetricWrapperBase(object):
             raise ValueError('No label names were set when constructing %s' % self)
 
         if self._labelvalues:
-            raise ValueError('%s already has labels set (%s); can not chain calls to .labels()' % (
+            raise ValueError('{} already has labels set ({}); can not chain calls to .labels()'.format(
                 self,
                 dict(zip(self._labelnames, self._labelvalues))
             ))
@@ -158,11 +180,11 @@ class MetricWrapperBase(object):
         if labelkwargs:
             if sorted(labelkwargs) != sorted(self._labelnames):
                 raise ValueError('Incorrect label names')
-            labelvalues = tuple(unicode(labelkwargs[l]) for l in self._labelnames)
+            labelvalues = tuple(str(labelkwargs[l]) for l in self._labelnames)
         else:
             if len(labelvalues) != len(self._labelnames):
                 raise ValueError('Incorrect label count')
-            labelvalues = tuple(unicode(l) for l in labelvalues)
+            labelvalues = tuple(str(l) for l in labelvalues)
         with self._lock:
             if labelvalues not in self._metrics:
                 self._metrics[labelvalues] = self.__class__(
@@ -175,37 +197,37 @@ class MetricWrapperBase(object):
                 )
             return self._metrics[labelvalues]
 
-    def remove(self, *labelvalues):
+    def remove(self, *labelvalues: Any) -> None:
         if not self._labelnames:
             raise ValueError('No label names were set when constructing %s' % self)
 
         """Remove the given labelset from the metric."""
         if len(labelvalues) != len(self._labelnames):
             raise ValueError('Incorrect label count (expected %d, got %s)' % (len(self._labelnames), labelvalues))
-        labelvalues = tuple(unicode(l) for l in labelvalues)
+        labelvalues = tuple(str(l) for l in labelvalues)
         with self._lock:
             del self._metrics[labelvalues]
 
-    def clear(self):
+    def clear(self) -> None:
         """Remove all labelsets from the metric"""
         with self._lock:
             self._metrics = {}
 
-    def _samples(self):
+    def _samples(self) -> Iterable[Sample]:
         if self._is_parent():
             return self._multi_samples()
         else:
             return self._child_samples()
 
-    def _multi_samples(self):
+    def _multi_samples(self) -> Iterable[Sample]:
         with self._lock:
             metrics = self._metrics.copy()
         for labels, metric in metrics.items():
             series_labels = list(zip(self._labelnames, labels))
-            for suffix, sample_labels, value in metric._samples():
-                yield (suffix, dict(series_labels + list(sample_labels.items())), value)
+            for suffix, sample_labels, value, timestamp, exemplar in metric._samples():
+                yield Sample(suffix, dict(series_labels + list(sample_labels.items())), value, timestamp, exemplar)
 
-    def _child_samples(self):  # pragma: no cover
+    def _child_samples(self) -> Iterable[Sample]:  # pragma: no cover
         raise NotImplementedError('_child_samples() must be implemented by %r' % self)
 
     def _metric_init(self):  # pragma: no cover
@@ -251,18 +273,22 @@ class Counter(MetricWrapperBase):
     """
     _type = 'counter'
 
-    def _metric_init(self):
+    def _metric_init(self) -> None:
         self._value = values.ValueClass(self._type, self._name, self._name + '_total', self._labelnames,
                                         self._labelvalues)
         self._created = time.time()
 
-    def inc(self, amount=1):
+    def inc(self, amount: float = 1, exemplar: Optional[Dict[str, str]] = None) -> None:
         """Increment counter by the given amount."""
+        self._raise_if_not_observable()
         if amount < 0:
             raise ValueError('Counters can only be incremented by non-negative amounts.')
         self._value.inc(amount)
+        if exemplar:
+            _validate_exemplar(exemplar)
+            self._value.set_exemplar(Exemplar(exemplar, amount, time.time()))
 
-    def count_exceptions(self, exception=Exception):
+    def count_exceptions(self, exception: Type[BaseException] = Exception) -> ExceptionCounter:
         """Count exceptions in a block of code or function.
 
         Can be used as a function decorator or context manager.
@@ -272,11 +298,14 @@ class Counter(MetricWrapperBase):
         self._raise_if_not_observable()
         return ExceptionCounter(self, exception)
 
-    def _child_samples(self):
-        return (
-            ('_total', {}, self._value.get()),
-            ('_created', {}, self._created),
-        )
+    def _child_samples(self) -> Iterable[Sample]:
+        sample = Sample('_total', {}, self._value.get(), None, self._value.get_exemplar())
+        if _use_created:
+            return (
+                sample,
+                Sample('_created', {}, self._created, None, None)
+            )
+        return (sample,)
 
 
 class Gauge(MetricWrapperBase):
@@ -320,20 +349,20 @@ class Gauge(MetricWrapperBase):
     _MULTIPROC_MODES = frozenset(('min', 'max', 'livesum', 'liveall', 'all'))
 
     def __init__(self,
-                 name,
-                 documentation,
-                 labelnames=(),
-                 namespace='',
-                 subsystem='',
-                 unit='',
-                 registry=REGISTRY,
-                 _labelvalues=None,
-                 multiprocess_mode='all',
+                 name: str,
+                 documentation: str,
+                 labelnames: Iterable[str] = (),
+                 namespace: str = '',
+                 subsystem: str = '',
+                 unit: str = '',
+                 registry: Optional[CollectorRegistry] = REGISTRY,
+                 _labelvalues: Optional[Sequence[str]] = None,
+                 multiprocess_mode: str = 'all',
                  ):
         self._multiprocess_mode = multiprocess_mode
         if multiprocess_mode not in self._MULTIPROC_MODES:
             raise ValueError('Invalid multiprocess mode: ' + multiprocess_mode)
-        super(Gauge, self).__init__(
+        super().__init__(
             name=name,
             documentation=documentation,
             labelnames=labelnames,
@@ -345,29 +374,32 @@ class Gauge(MetricWrapperBase):
         )
         self._kwargs['multiprocess_mode'] = self._multiprocess_mode
 
-    def _metric_init(self):
+    def _metric_init(self) -> None:
         self._value = values.ValueClass(
             self._type, self._name, self._name, self._labelnames, self._labelvalues,
             multiprocess_mode=self._multiprocess_mode
         )
 
-    def inc(self, amount=1):
+    def inc(self, amount: float = 1) -> None:
         """Increment gauge by the given amount."""
+        self._raise_if_not_observable()
         self._value.inc(amount)
 
-    def dec(self, amount=1):
+    def dec(self, amount: float = 1) -> None:
         """Decrement gauge by the given amount."""
+        self._raise_if_not_observable()
         self._value.inc(-amount)
 
-    def set(self, value):
+    def set(self, value: float) -> None:
         """Set gauge to the given value."""
+        self._raise_if_not_observable()
         self._value.set(float(value))
 
-    def set_to_current_time(self):
+    def set_to_current_time(self) -> None:
         """Set gauge to the current unixtime."""
         self.set(time.time())
 
-    def track_inprogress(self):
+    def track_inprogress(self) -> InprogressTracker:
         """Track inprogress blocks of code or functions.
 
         Can be used as a function decorator or context manager.
@@ -377,28 +409,29 @@ class Gauge(MetricWrapperBase):
         self._raise_if_not_observable()
         return InprogressTracker(self)
 
-    def time(self):
+    def time(self) -> Timer:
         """Time a block of code or function, and set the duration in seconds.
 
         Can be used as a function decorator or context manager.
         """
-        self._raise_if_not_observable()
-        return Timer(self.set)
+        return Timer(self, 'set')
 
-    def set_function(self, f):
+    def set_function(self, f: Callable[[], float]) -> None:
         """Call the provided function to return the Gauge value.
 
         The function must return a float, and may be called from
         multiple threads. All other methods of the Gauge become NOOPs.
         """
 
-        def samples(self):
-            return (('', {}, float(f())),)
+        self._raise_if_not_observable()
 
-        self._child_samples = create_bound_method(samples, self)
+        def samples(_: Gauge) -> Iterable[Sample]:
+            return (Sample('', {}, float(f()), None, None),)
 
-    def _child_samples(self):
-        return (('', {}, self._value.get()),)
+        self._child_samples = types.MethodType(samples, self)  # type: ignore
+
+    def _child_samples(self) -> Iterable[Sample]:
+        return (Sample('', {}, self._value.get(), None, None),)
 
 
 class Summary(MetricWrapperBase):
@@ -434,13 +467,13 @@ class Summary(MetricWrapperBase):
     _type = 'summary'
     _reserved_labelnames = ['quantile']
 
-    def _metric_init(self):
+    def _metric_init(self) -> None:
         self._count = values.ValueClass(self._type, self._name, self._name + '_count', self._labelnames,
                                         self._labelvalues)
         self._sum = values.ValueClass(self._type, self._name, self._name + '_sum', self._labelnames, self._labelvalues)
         self._created = time.time()
 
-    def observe(self, amount):
+    def observe(self, amount: float) -> None:
         """Observe the given amount.
 
         The amount is usually positive or zero. Negative values are
@@ -450,22 +483,25 @@ class Summary(MetricWrapperBase):
         https://prometheus.io/docs/practices/histograms/#count-and-sum-of-observations
         for details.
         """
+        self._raise_if_not_observable()
         self._count.inc(1)
         self._sum.inc(amount)
 
-    def time(self):
+    def time(self) -> Timer:
         """Time a block of code or function, and observe the duration in seconds.
 
         Can be used as a function decorator or context manager.
         """
-        self._raise_if_not_observable()
-        return Timer(self.observe)
+        return Timer(self, 'observe')
 
-    def _child_samples(self):
-        return (
-            ('_count', {}, self._count.get()),
-            ('_sum', {}, self._sum.get()),
-            ('_created', {}, self._created))
+    def _child_samples(self) -> Iterable[Sample]:
+        samples = [
+            Sample('_count', {}, self._count.get(), None, None),
+            Sample('_sum', {}, self._sum.get(), None, None),
+        ]
+        if _use_created:
+            samples.append(Sample('_created', {}, self._created, None, None))
+        return tuple(samples)
 
 
 class Histogram(MetricWrapperBase):
@@ -508,18 +544,18 @@ class Histogram(MetricWrapperBase):
     DEFAULT_BUCKETS = (.005, .01, .025, .05, .075, .1, .25, .5, .75, 1.0, 2.5, 5.0, 7.5, 10.0, INF)
 
     def __init__(self,
-                 name,
-                 documentation,
-                 labelnames=(),
-                 namespace='',
-                 subsystem='',
-                 unit='',
-                 registry=REGISTRY,
-                 _labelvalues=None,
-                 buckets=DEFAULT_BUCKETS,
+                 name: str,
+                 documentation: str,
+                 labelnames: Iterable[str] = (),
+                 namespace: str = '',
+                 subsystem: str = '',
+                 unit: str = '',
+                 registry: Optional[CollectorRegistry] = REGISTRY,
+                 _labelvalues: Optional[Sequence[str]] = None,
+                 buckets: Sequence[Union[float, str]] = DEFAULT_BUCKETS,
                  ):
         self._prepare_buckets(buckets)
-        super(Histogram, self).__init__(
+        super().__init__(
             name=name,
             documentation=documentation,
             labelnames=labelnames,
@@ -531,8 +567,8 @@ class Histogram(MetricWrapperBase):
         )
         self._kwargs['buckets'] = buckets
 
-    def _prepare_buckets(self, buckets):
-        buckets = [float(b) for b in buckets]
+    def _prepare_buckets(self, source_buckets: Sequence[Union[float, str]]) -> None:
+        buckets = [float(b) for b in source_buckets]
         if buckets != sorted(buckets):
             # This is probably an error on the part of the user,
             # so raise rather than sorting for them.
@@ -543,8 +579,8 @@ class Histogram(MetricWrapperBase):
             raise ValueError('Must have at least two buckets')
         self._upper_bounds = buckets
 
-    def _metric_init(self):
-        self._buckets = []
+    def _metric_init(self) -> None:
+        self._buckets: List[values.ValueClass] = []
         self._created = time.time()
         bucket_labelnames = self._labelnames + ('le',)
         self._sum = values.ValueClass(self._type, self._name, self._name + '_sum', self._labelnames, self._labelvalues)
@@ -557,7 +593,7 @@ class Histogram(MetricWrapperBase):
                 self._labelvalues + (floatToGoString(b),))
             )
 
-    def observe(self, amount):
+    def observe(self, amount: float, exemplar: Optional[Dict[str, str]] = None) -> None:
         """Observe the given amount.
 
         The amount is usually positive or zero. Negative values are
@@ -567,29 +603,34 @@ class Histogram(MetricWrapperBase):
         https://prometheus.io/docs/practices/histograms/#count-and-sum-of-observations
         for details.
         """
+        self._raise_if_not_observable()
         self._sum.inc(amount)
         for i, bound in enumerate(self._upper_bounds):
             if amount <= bound:
                 self._buckets[i].inc(1)
+                if exemplar:
+                    _validate_exemplar(exemplar)
+                    self._buckets[i].set_exemplar(Exemplar(exemplar, amount, time.time()))
                 break
 
-    def time(self):
+    def time(self) -> Timer:
         """Time a block of code or function, and observe the duration in seconds.
 
         Can be used as a function decorator or context manager.
         """
-        return Timer(self.observe)
+        return Timer(self, 'observe')
 
-    def _child_samples(self):
+    def _child_samples(self) -> Iterable[Sample]:
         samples = []
-        acc = 0
+        acc = 0.0
         for i, bound in enumerate(self._upper_bounds):
             acc += self._buckets[i].get()
-            samples.append(('_bucket', {'le': floatToGoString(bound)}, acc))
-        samples.append(('_count', {}, acc))
+            samples.append(Sample('_bucket', {'le': floatToGoString(bound)}, acc, None, self._buckets[i].get_exemplar()))
+        samples.append(Sample('_count', {}, acc, None, None))
         if self._upper_bounds[0] >= 0:
-            samples.append(('_sum', {}, self._sum.get()))
-        samples.append(('_created', {}, self._created))
+            samples.append(Sample('_sum', {}, self._sum.get(), None, None))
+        if _use_created:
+            samples.append(Sample('_created', {}, self._created, None, None))
         return tuple(samples)
 
 
@@ -616,17 +657,17 @@ class Info(MetricWrapperBase):
         self._lock = Lock()
         self._value = {}
 
-    def info(self, val):
+    def info(self, val: Dict[str, str]) -> None:
         """Set info metric."""
         if self._labelname_set.intersection(val.keys()):
-            raise ValueError('Overlapping labels for Info metric, metric: %s child: %s' % (
+            raise ValueError('Overlapping labels for Info metric, metric: {} child: {}'.format(
                 self._labelnames, val))
         with self._lock:
             self._value = dict(val)
 
-    def _child_samples(self):
+    def _child_samples(self) -> Iterable[Sample]:
         with self._lock:
-            return (('_info', self._value, 1.0,),)
+            return (Sample('_info', self._value, 1.0, None, None),)
 
 
 class Enum(MetricWrapperBase):
@@ -645,17 +686,17 @@ class Enum(MetricWrapperBase):
     _type = 'stateset'
 
     def __init__(self,
-                 name,
-                 documentation,
-                 labelnames=(),
-                 namespace='',
-                 subsystem='',
-                 unit='',
-                 registry=REGISTRY,
-                 _labelvalues=None,
-                 states=None,
+                 name: str,
+                 documentation: str,
+                 labelnames: Sequence[str] = (),
+                 namespace: str = '',
+                 subsystem: str = '',
+                 unit: str = '',
+                 registry: Optional[CollectorRegistry] = REGISTRY,
+                 _labelvalues: Optional[Sequence[str]] = None,
+                 states: Optional[Sequence[str]] = None,
                  ):
-        super(Enum, self).__init__(
+        super().__init__(
             name=name,
             documentation=documentation,
             labelnames=labelnames,
@@ -666,25 +707,25 @@ class Enum(MetricWrapperBase):
             _labelvalues=_labelvalues,
         )
         if name in labelnames:
-            raise ValueError('Overlapping labels for Enum metric: %s' % (name,))
+            raise ValueError(f'Overlapping labels for Enum metric: {name}')
         if not states:
-            raise ValueError('No states provided for Enum metric: %s' % (name,))
+            raise ValueError(f'No states provided for Enum metric: {name}')
         self._kwargs['states'] = self._states = states
 
-    def _metric_init(self):
+    def _metric_init(self) -> None:
         self._value = 0
         self._lock = Lock()
 
-    def state(self, state):
+    def state(self, state: str) -> None:
         """Set enum metric state."""
         self._raise_if_not_observable()
         with self._lock:
             self._value = self._states.index(state)
 
-    def _child_samples(self):
+    def _child_samples(self) -> Iterable[Sample]:
         with self._lock:
             return [
-                ('', {self._name: s}, 1 if i == self._value else 0,)
+                Sample('', {self._name: s}, 1 if i == self._value else 0, None, None)
                 for i, s
                 in enumerate(self._states)
             ]
